# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# Copyright (c) 2018 IRT-AESE.
# All rights reserved.
# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: François Gallard: initial Scilab version
#        :author: Arthur Piat: conversion Scilab to Matlab and complementary features
#        :author: Nicolas Roussouly: GEMSEO integration
"""Definition of the Matlab discipline.

Overview
--------

This module contains the :class:`.MatlabDiscipline`
which enables to automatically create a wrapper of any Matlab function.
This class can be used in order to interface any Matlab code
and to use it inside a MDO process.
"""
from __future__ import annotations

import logging
import os
import re
from os.path import exists
from os.path import join
from pathlib import Path
from typing import Any
from typing import ClassVar
from typing import Mapping
from typing import MutableMapping
from typing import Sequence

import matlab.engine
import numpy as np
from gemseo.core.discipline import MDODiscipline

from gemseo_matlab.engine import get_matlab_engine
from gemseo_matlab.engine import MatlabEngine
from gemseo_matlab.matlab_data_processor import convert_array_from_matlab
from gemseo_matlab.matlab_data_processor import double2array
from gemseo_matlab.matlab_data_processor import load_matlab_file
from gemseo_matlab.matlab_data_processor import MatlabDataProcessor
from gemseo_matlab.matlab_data_processor import save_matlab_file
from gemseo_matlab.matlab_parser import MatlabParser

LOGGER = logging.getLogger(__name__)


class MatlabDiscipline(MDODiscipline):
    """Base wrapper for matlab discipline.

    Generates a discipline of given matlab function and wrap it to be executed with
    GEMSEO. Can be used on encrypted, MATLAB build-in and user made function.

    Examples:
        >>> # build the discipline from the MATLAB function "function.m"
        >>> disc = MatlabDiscipline("function.m")
        >>> # Execute the discipline
        >>> disc.execute({"x" : array([2.]), "y" : array([1.])})
        >>>
        >>> # build discipline with initial data from MATLAB file
        >>> disc = MatlabDiscipline("function.m", matlab_data_file="data.mat")
        >>> # execute discipline from default values
        >>> disc.execute()
        >>>
        >>> # build discipline from MATLAB file located in matlab_files directory
        >>> disc = MatlabDiscipline("function.m", search_file="matlab_files")
        >>>
        >>> # build discipline with jacobian returned by the matlab function
        >>> disc = MatlabDiscipline("function.m", is_jac_returned_by_func=True)
        >>> disc.execute({"x" : array([2.]), "y" : array([1.])})
        >>> # print jacboian values
        >>> print(disc.jac)

    Note:
        If ``is_jac_returned_by_func`` is True, jacobian matrices must be returned
        by the matlab function itself. In such case, function outputs must contain
        standard output as well as new outputs for jacobian terms. These new
        outputs must follow naming convention described in function
        :meth:`.MatlabDiscipline._get_jac_name`. They can be returned
        in any order.
    """

    JAC_PREFIX: ClassVar[str] = "jac_"

    def __init__(
        self,
        matlab_fct: str | Path,
        input_names: Sequence[str] | None = None,
        output_names: Sequence[str] | None = None,
        add_subfold_path: bool = False,
        search_file: str | None = None,
        matlab_engine_name: str = "matlab",
        matlab_data_file: str | Path | None = None,
        name: str | None = None,
        clean_cache_each_n: int | None = None,
        input_grammar_file: str | None = None,
        output_grammar_file: str | None = None,
        auto_detect_grammar_files: bool = False,
        check_opt_data: bool = True,
        cache_type: MDODiscipline.CacheType = MDODiscipline.CacheType.SIMPLE,
        grammar_type: MDODiscipline.GrammarType = MDODiscipline.GrammarType.JSON,
        cache_file_path: str | None = None,
        is_jac_returned_by_func: bool = False,
    ) -> None:
        """
        Args:
            matlab_fct: The path of the Matlab file or Name of the function.
            input_names: The input variables.
            output_names: The output variables.
            add_subfold_path: Whether to add all sub-folder to matlab engine path.
            search_file: The root directory to launch the research of matlab file.
            matlab_engine_name: The name of the singleton used for this discipline.
            matlab_data_file: The .mat file or path containing default values of data.
            name: The name of discipline.
            clean_cache_each_n: Iteration interval at which matlab workspace is cleaned.
            input_grammar_file: The file for input grammar description,
                if None, name + "_input.json" is used.
            output_grammar_file: The file for output grammar description.
            auto_detect_grammar_files: If no input and output grammar files
                are provided,
                auto_detect_grammar_files uses a naming convention
                to associate a grammar file to a discipline:
                searches in the "comp_dir" directory containing the
                discipline source file for files basenames
                self.name _input.json and self.name _output.json.
            check_opt_data: Whether to check input and output data of
                discipline.
            cache_type: The type of cache.
            grammar_type: The type of the input and output grammars.
            cache_file_path: The file to store the data,
                mandatory when HDF caching is used.
            is_jac_returned_by_func: If True, the jacobian matrices should be returned
                of matlab function with standard outputs.
                Default is False.
                If True, the conventional name 'jac_dout_din' is used as jacobian
                term of any output 'out' with respect to input 'in'.
        """  # noqa: D205, D212, D415
        super().__init__(
            name=name,
            input_grammar_file=input_grammar_file,
            output_grammar_file=output_grammar_file,
            auto_detect_grammar_files=auto_detect_grammar_files,
            cache_type=cache_type,
            grammar_type=grammar_type,
            cache_file_path=cache_file_path,
        )
        self.__fct_name = None

        matlab_fct = str(matlab_fct)
        if input_names is None or output_names is None:
            parser = MatlabParser()

            if search_file is not None:
                path = self.search_file(matlab_fct, search_file)
                parser.parse(path)
                if matlab_data_file is not None and not exists(str(matlab_data_file)):
                    matlab_data_file = self.search_file(
                        str(matlab_data_file), search_file, ".mat"
                    )
            else:
                parser.parse(matlab_fct)

            input_data = parser.inputs
            output_data = parser.outputs
            function_path = (parser.directory / parser.function_name).with_suffix(".m")
        else:
            function_path = matlab_fct

        if input_names is not None:
            input_data = input_names
        if output_names is not None:
            output_data = output_names

        self.__engine = get_matlab_engine(matlab_engine_name)
        self.__inputs = input_data
        self.__outputs = output_data
        # init size with -1 -> means that size is currently unknown
        self.__is_size_known = False
        self.__inputs_size = {var: -1 for var in self.__inputs}
        self.__outputs_size = {var: -1 for var in self.__outputs}

        # self.outputs can be filtered here
        self.__is_jac_returned_by_func = is_jac_returned_by_func
        self.__jac_output_names = []
        self.__jac_output_indices = []
        if self.__is_jac_returned_by_func:
            self.__filter_jacobian_in_outputs()

        self.__check_function(function_path, add_subfold_path)
        self.__check_opt_data = check_opt_data
        self.cleaning_interval = clean_cache_each_n
        self.__init_default_data(
            matlab_data_file,
            input_grammar_file,
            output_grammar_file,
            auto_detect_grammar_files,
        )
        self.data_processor = MatlabDataProcessor()

        if self.__is_jac_returned_by_func:
            self.__reorder_and_check_jacobian_consistency()

    @property
    def engine(self) -> MatlabEngine:
        """The matlab engine of the discipline.

        The engine is associated to the ``matlab_engine_name`` provided at the instance
        construction.
        """
        return self.__engine

    @property
    def function_name(self) -> str:
        """Return the name of the function."""
        return self.__fct_name

    @staticmethod
    def search_file(
        file_name: str,
        root_dir: str,
        extension: str = ".m",
    ) -> str:
        """Locate recursively a file in the given root directory.

        Args:
            file_name: The name of the file to be located.
            root_dir: The root directory to launch the research.
            extension: The extension of the file in case not given by user.

        Returns:
            The path of the given file.

        Raises:
            IOError:
                * If two files are found in same directory;
                * If no file is found.
        """
        found_file = False
        re_matfile = re.compile(r"\S+\.\S*")
        grps = re_matfile.search(file_name)
        if grps is None:
            file_name += extension

        file_path = ""
        for subdir, _, files in os.walk(str(root_dir)):
            for file_loc in files:
                if file_loc == file_name:
                    if found_file:
                        msg = "At least two files {} were in directory {}".format(
                            file_name, root_dir
                        )
                        msg += f"\n File one: {file_path};"
                        msg += f"\n File two: {join(subdir, file_loc)}."
                        raise OSError(msg)
                    found_file = True
                    file_path = join(subdir, file_loc)
                    dir_name = subdir

        if not found_file:
            raise OSError(f"No file: {file_name}, found in directory: {root_dir}.")

        LOGGER.info("File: %s found in directory: %s.", file_name, dir_name)
        return file_path

    def __check_function(
        self,
        matlab_fct: str | Path,
        add_subfold_path: bool,
    ) -> None:
        """Check the availability of the prescribed MATLAB function.

        The function manages encrypted, build-in and user made function and
        unify their use.

        Args:
            matlab_fct: A name for the matlab function to be wrapped.
            add_subfold_path: If true, add all sub-folders of the function to
                matlab search path.

        Raises:
            NameError: If the function (or file) does not exist.
        """
        path = Path(matlab_fct)
        if path.exists():
            # Test if the file exists in the system
            self.__engine.add_path(path.parent, add_subfolder=add_subfold_path)
            self.__fct_name = path.stem
        elif self.__engine.exist(matlab_fct)[0]:
            # If file does not exist, try to find an existing build-in function in
            # engine
            self.__fct_name = matlab_fct
        else:
            # If no file and build-in function exist, raise error
            msg = f'No existing file or function "{matlab_fct}".'
            raise NameError(msg)

    def __init_default_data(
        self,
        matlab_data_file: str,
        input_grammar_file: str,
        output_grammar_file: str,
        auto_detect_grammar_files: bool,
    ) -> None:
        """Initialize default data of the discipline.

        Args:
            matlab_data_file: The path to the .mat containing default values of data
            input_grammar_file: The file for input grammar description,
                if None, name + "_input.json" is used.
            output_grammar_file: The file for output grammar description.
            auto_detect_grammar_files: If True, no input and output grammar files.
        """
        if matlab_data_file is not None:
            saved_values = convert_array_from_matlab(
                load_matlab_file(str(matlab_data_file).replace(".mat", ""))
            )

        if input_grammar_file is None and not auto_detect_grammar_files:
            # Here, we temporary init inputs data with an array of
            # size 1 but that could not be the right size...
            # The right size can be known from either matlab_data_file or evaluating
            # the matlab function
            input_data = dict.fromkeys(self.__inputs, np.array([0.1]))
            if matlab_data_file is not None:
                input_data = self.__update_data(input_data.copy(), saved_values)

        if output_grammar_file is None and not auto_detect_grammar_files:
            # same remark as above about the size
            output_data = dict.fromkeys(self.__outputs, np.array([0.1]))
            if matlab_data_file is not None:
                output_data = self.__update_data(output_data.copy(), saved_values)

        self.input_grammar.update_from_data(input_data)

        self.output_grammar.update_from_data(output_data)

        # If none input matlab data is prescribed, we cannot know
        # the size of inputs and outputs. Thus, we must evaluate
        # the function in order to know the sizes
        if matlab_data_file is not None:
            self.__is_size_known = True
            for input_name, input_value in input_data.items():
                self.__inputs_size[input_name] = len(input_value)

            for output_name, output_value in output_data.items():
                self.__outputs_size[output_name] = len(output_value)

        self.default_inputs = input_data.copy()

    def __filter_jacobian_in_outputs(self) -> None:
        """Filter jacobians in outputs names.

        This function is applied when _is_jac_returned_by_func is True. In such case,
        the function extracts the jacobian component from the list of output names
        returned by the matlab function. It thus fills _jac_output_names attributes as
        well as _jac_output_indices which corresponds to indices of jacobian component
        in the list of outputs returned by the matlab function.

        After applying this function, _outputs attribute no longer contains jacobian
        output names but only standard outputs.

        In order to filter jacobian component, this function just checks that jacobian
        names are prefixed by 'jac_'.
        """
        output_names = list(self.__outputs)

        # select jacobian output and remove them from self.outputs
        for i, out_name in enumerate(self.__outputs):
            if out_name[0:4] == self.JAC_PREFIX:
                self.__jac_output_names.append(out_name)
                self.__jac_output_indices.append(i)
                output_names.remove(out_name)

        # here self.outputs only contains output responses (no jacobian)
        self.__outputs = output_names

    def __reorder_and_check_jacobian_consistency(self) -> None:
        """This function checks jacobian output consistency.

        This function is used when _is_jac_returned_by_func is True.

        The function is called after calling jacobian filtering
        :meth:`.MatlabDiscipline._filter_jacobian_in_outputs`. It enables to:
        * check that all outputs have a jacobian matrix with respect to all inputs;
        * reorder the list of jacobian names (and indices) following the order
          from iterating over outputs then inputs lists;

        In order to check that all jacobian components exist, the function
        uses the conventional naming described in
        :meth:`.MatlabDiscipline._get_jac_name`.

        Raises:
            ValueError:
                * If the number of jacobian outputs is wrong;
                * If a specific jacobian output has the wrong name.
        """
        conventional_jac_names = self.__get_conventional_jac_names()
        new_indices = [-1] * len(conventional_jac_names)

        if len(conventional_jac_names) != len(self.__jac_output_names):
            raise ValueError(
                "The number of jacobian outputs does "
                "not correspond to what it should be. "
                "Make sure that all outputs have a jacobian "
                "matrix with respect to inputs."
            )

        not_found = []
        for i, name in enumerate(conventional_jac_names):
            try:
                idx = self.__jac_output_names.index(name)
                new_indices[i] = self.__jac_output_indices[idx]
            except ValueError:
                not_found.append(name)

        if not_found:
            raise ValueError(
                "Jacobian terms {} are not found in the "
                "list of conventional names. It is reminded that "
                "jacobian terms' name should be "
                "such as 'jac_dout_din'".format(not_found)
            )

        self.__jac_output_names = conventional_jac_names
        self.__jac_output_indices = new_indices

    def __get_conventional_jac_names(self) -> list[str]:
        """Return the list of jacobian names following the conventional naming.

        The conventional naming is described in :meth:`.MatlabDiscipline._get_jac_name`.
        """
        return [
            self.__get_jac_name(out_var, in_var)
            for out_var in self.__outputs
            for in_var in self.__inputs
        ]

    def __get_jac_name(
        self,
        out_var: str,
        in_var: str,
    ) -> str:
        """Return the name of jacobian given input and ouput variables.

        The conventional naming of jacobian component is the following:
        if outputs have any names ``out_1``, ``out_2``...
        and inputs are ``in_1``, ``in_2``... Therefore, names of jacobian
        components returned
        by the matlab function must be: ``jac_dout_1_din_1``,
        ``jac_dout_1_din_2``,
        ``jac_dout_2_din_1``, ``jac_dout_2_din_2``... which means
        that the names must be prefixed by ``jac_``, and followed by
        ``doutput`` and ``dinput`` seperated by ``_``.

        Args:
            out_var: The output variable name.
            in_var: The input variable name.

        Returns:
            The jacobian matrix name of output with respect to input.
        """
        return str(
            "{prefix}d{outv}_d{inv}".format(
                prefix=self.JAC_PREFIX, outv=out_var, inv=in_var
            )
        )

    def check_input_data(  # noqa: D102
        self,
        input_data: Mapping[str, np.ndarray],
        raise_exception: bool = True,
    ) -> None:
        if self.__check_opt_data:
            super().check_input_data(input_data, raise_exception=raise_exception)

    def check_output_data(self, raise_exception: bool = True) -> None:  # noqa: D102
        if self.__check_opt_data:
            super().check_output_data(raise_exception=raise_exception)

    def _run(self) -> None:
        """Run the Matlab discipline.

        If jacobian values are returned by the matlab function, they are filtered and
        used in order to fill :attr:`.MatlabDiscipline.jac`.

        Raises:
            ValueError:
                * If the execution of the matlab function fails.
                * If the size of the jacobian output matrix is wrong.
        """
        # import pudb;pudb.set_trace()
        input_vals = self.get_input_data()
        list_of_values = [input_vals.get(k) for k in self.__inputs if k in input_vals]

        try:
            out_vals = self.__engine.execute_function(
                self.__fct_name,
                *list_of_values,
                nargout=len(self.__outputs) + len(self.__jac_output_names),
            )

        except matlab.engine.MatlabExecutionError:
            LOGGER.error("Discipline: %s execution failed", self.name)
            raise

        # filter output values if jacobian is returned
        jac_vals = []

        if self.__is_jac_returned_by_func:
            out_vals = np.array(out_vals, dtype=object)
            jac_vals = [out_vals[idx] for idx in self.__jac_output_indices]
            out_vals = np.delete(out_vals, self.__jac_output_indices)
            # --> now out_vals only contains output responses (no jacobian)

        if self.cleaning_interval is not None:
            if self._n_calls.value % self.cleaning_interval == 0:
                self.__engine.execute_function("clear", "all", nargout=0)
                LOGGER.info(
                    "MATLAB cache cleaned: Discipline called %s times",
                    self._n_calls.value,
                )

        out_names = self.__outputs

        if len(out_names) == 1:
            self.store_local_data(**{out_names[0]: double2array(out_vals)})
        else:
            for out_n, out_v in zip(out_names, out_vals):
                self.store_local_data(**{out_n: double2array(out_v)})

        if not self.__is_size_known:
            for i, var in enumerate(self.__inputs):
                self.__inputs_size[var] = len(list_of_values[i])
            for var in self.__outputs:
                self.__outputs_size[var] = len(self.local_data[var])
            self.__is_size_known = True

        if self.__is_jac_returned_by_func:
            # fill jac dict
            self._init_jacobian()
            cpt = 0
            for out_name in self.__outputs:
                self.jac[out_name] = {}
                for in_name in self.__inputs:
                    self.jac[out_name][in_name] = np.atleast_2d(jac_vals[cpt])

                    if self.jac[out_name][in_name].shape != (
                        self.__outputs_size[out_name],
                        self.__inputs_size[in_name],
                    ):
                        raise ValueError(
                            "Jacobian term 'jac_d{}_d{}' "
                            "has the wrong size {} whereas it should "
                            "be {}.".format(
                                out_name,
                                in_name,
                                self.jac[out_name][in_name].shape,
                                (
                                    self.__outputs_size[out_name],
                                    self.__inputs_size[in_name],
                                ),
                            )
                        )

                    cpt += 1

            self._is_linearized = True

    @staticmethod
    def __update_data(
        data: MutableMapping[str, Any],
        other_data: Mapping[str, Any],
    ) -> Mapping[str, Any]:
        """Update the values of a data mapping without adding new data names.

        Args:
            data: The data to be updated.
            other_data: The data to update ``data``.

        Returns:
            The updated data.
        """
        for key, value in other_data.items():
            if key in data.keys():
                data[key] = value

        return data

    def save_data_to_matlab(self, file_path: str | Path) -> None:
        """Save local data to matlab .mat format.

        Args:
            file_path: The path where to save the file.
        """
        file_path = Path(file_path)
        save_matlab_file(self.local_data, file_path=file_path)
        msg = "Local data of discipline {} exported to {}.mat successfully.".format(
            self.name, file_path.name
        )
        LOGGER.info(msg)

    @property
    def cleaning_interval(self) -> int:
        """Get and/or set the flushing interval for matlab disciplines."""
        return self.__cleaning_interval

    @cleaning_interval.setter
    def cleaning_interval(self, cleaning_interval: int) -> None:
        if cleaning_interval is not None:
            is_integer = cleaning_interval % 1 == 0
            if not is_integer:
                raise ValueError(
                    "The parameter 'cleaning_interval' argument must be an integer."
                )
        self.__cleaning_interval = cleaning_interval



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# Copyright (c) 2018 IRT-AESE.
# All rights reserved.
# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: François Gallard : initial Scilab version
#        :author: Arthur Piat : conversion Scilab to Matlab
#        :author: Nicolas Roussouly: GEMSEO integration
"""Definition of the Matlab parser.

Overview
--------

This module contains the :class:`.MatlabParser`
which enables to parse Matlab files in order to automatically
detect inputs and outputs.
This class is basically used through the :class:`.MatlabDiscipline` class
in order to build a discipline based on the Matlab function.
"""
from __future__ import annotations

import logging
import re
from pathlib import Path

LOGGER = logging.getLogger(__name__)


class MatlabParser:
    """Parse Matlab file to identify inputs and outputs.

    Examples:
        >>> # Parse the matlab function "fucntion.m"
        >>> parser = MatlabParser("function.m")
        >>> print(parser.inputs)
        >>> print(parser.outputs)
    """

    RE_FILE_FMT = re.compile(r".*(\\.*)*\.m\b")
    RE_ENCRYPTED_FCT = re.compile(r".*(\\.*)*\.p\b")
    RE_OUTPUTS = re.compile(r"(\[(.*?)])|(function( *?)(.*?)=)")
    RE_FUNCTION = re.compile(r"=(.*?)\(")
    RE_ARGS = re.compile(r"\((.*?)\)")

    def __init__(self, full_path: str | None = None) -> None:
        """
        Args:
            full_path: The path to the matlab file.
                If ``None``, the user shall parse the file explicitly.
        """  # noqa: D205, D212, D415
        self.__inputs = None
        self.__outputs = None
        self.__fct_name = None
        self.__fct_dir = None

        if full_path is not None:
            self.parse(full_path)

    @property
    def function_name(self) -> str:
        """Return the name of the function."""
        return self.__fct_name

    @property
    def directory(self) -> Path:
        """Return the directory of the function."""
        return self.__fct_dir

    @property
    def inputs(self) -> list[str]:
        """Return the inputs."""
        return self.__inputs

    @property
    def outputs(self) -> list[str]:
        """Return the outputs."""
        return self.__outputs

    def __check_path(self, file_path: str | Path) -> None:
        """Check the format of the file.

        Args:
            file_path: The path to the matlab the file.

        Raises:
            FileNotFoundError: If the matlab file does not exist.
            ValueError:
                * If the matlab file is encrypted;
                * If the matlab function is neither a script nor a function.
        """
        file_path = Path(file_path)
        if not file_path.exists():
            raise FileNotFoundError(
                "The function directory for Matlab "
                f"sources {file_path} does not exists."
            )

        file_path = str(file_path)

        re_encrypted_file_groups = self.RE_ENCRYPTED_FCT.search(file_path)

        if re_encrypted_file_groups is not None:
            raise ValueError(
                f"The given file {file_path} is encrypted and cannot be parsed."
            )

        re_file_groups = self.RE_FILE_FMT.search(file_path)

        if re_file_groups is None:
            raise ValueError(
                f"The given file {file_path} should "
                "either be a matlab function or script."
            )

    def __parse_function_inputs_outputs(
        self,
        line: str,
        function_name: str,
    ) -> None:
        """Parse inputs and outputs.

        Args:
            line: The line containing the declaration of the function.
            function_name: The name of the function according to file.m.

        Raises:
            NameError:
                * If function has no name;
                * If function and file name are different.
            ValueError:
                * If function has no output;
                * If function has no input.
        """
        # TODO: refactor this function because some exceptions are not useful

        re_func_groups = self.RE_FUNCTION.search(line)

        if re_func_groups is None:
            raise NameError("Matlab function has no name.")

        fname = re_func_groups.group(0).strip()[1:-1].strip()

        if fname != function_name:
            raise NameError(
                "Function name {} does not match with file name {}.".format(
                    function_name, fname
                )
            )

        LOGGER.debug("Detected function: %s", fname)
        self.__fct_name = fname

        re_output_groups = self.RE_OUTPUTS.search(line)

        if re_output_groups is None:
            raise ValueError(f"Function {fname} has no output")

        arg_str = re_output_groups.group(0).strip()
        arg_str = arg_str.replace("[", "").replace("]", "")
        arg_str = arg_str.replace("function", "").replace(" ", "")
        arg_str = arg_str.replace("=", "")
        outs = arg_str.split(",")
        self.__outputs = [out_str.strip() for out_str in outs]
        LOGGER.debug("Outputs are: %s", outs)

        re_args_groups = self.RE_ARGS.search(line)
        if re_args_groups is None:
            raise ValueError(f"Function {fname} has no argument.")

        arg_str = re_args_groups.group(0).strip()[1:-1].strip()
        args = arg_str.split(",")
        self.__inputs = [args_str.strip() for args_str in args]
        LOGGER.debug("And arguments are: %s", args)

    def parse(self, path: str | Path) -> None:
        """Parse a .m file in order to get inputs and outputs.

        Args:
            path: The path to the matlab file.

        Raises:
            ValueError: Raised if the file is not a matlab function.
        """
        path = Path(path).absolute()
        self.__check_path(path)
        self.__fct_dir = path.parent
        fct_name = path.stem

        is_parsed = False

        with path.open(errors="ignore") as file_handle:
            for line in file_handle.readlines():
                if line.strip().startswith("function"):
                    self.__parse_function_inputs_outputs(line, fct_name)
                    is_parsed = True
                    break

        if not is_parsed:
            raise ValueError(f"The given file {path} is not a matlab function.")



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
from __future__ import annotations

import pytest
from gemseo_matlab.matlab_data_processor import load_matlab_file  # noqa: E402
from gemseo_matlab.matlab_discipline import MatlabDiscipline  # noqa: E402
from numpy import array

from .matlab_files import MATLAB_FILES_DIR_PATH  # noqa: E402

MATLAB_SIMPLE_FUNC = MATLAB_FILES_DIR_PATH / "dummy_test.m"
MATLAB_COMPLEX_FUNC = MATLAB_FILES_DIR_PATH / "dummy_complex_fct.m"
MATLAB_SIMPLE_FUNC_MULTIDIM = MATLAB_FILES_DIR_PATH / "dummy_test_multidim.m"
MATLAB_SIMPLE_FUNC_MULTIDIM_JAC = MATLAB_FILES_DIR_PATH / "dummy_test_multidim_jac.m"
FCT_MULTIDIM_DATASET = MATLAB_FILES_DIR_PATH / "dummy_file_multidim_fct.mat"


def test_engine_property():
    """Check the engine property."""
    mat = MatlabDiscipline(MATLAB_SIMPLE_FUNC)
    assert not mat.engine.is_closed


def test_inputs_from_matlab():
    """Test input variables read from matlab file."""
    mat2 = MatlabDiscipline(MATLAB_COMPLEX_FUNC)
    assert mat2._MatlabDiscipline__inputs == ["a", "b", "c", "d", "e", "f"]


def test_inputs_from_param():
    """Test input variables given as input param."""
    mat = MatlabDiscipline(
        MATLAB_COMPLEX_FUNC, input_names=["v1", "v2", "v3", "v4", "v5", "v6"]
    )
    assert mat._MatlabDiscipline__inputs == ["v1", "v2", "v3", "v4", "v5", "v6"]


def test_outputs():
    """Test output variables."""
    mat2 = MatlabDiscipline(MATLAB_COMPLEX_FUNC)
    assert mat2._MatlabDiscipline__outputs == ["x", "y", "z"]


def test_inputs_and_outputs_size_unknown():
    """Test that size of input and output variables are unknown when initializing
    without matlab data."""
    mat1 = MatlabDiscipline(MATLAB_SIMPLE_FUNC)

    assert mat1._MatlabDiscipline__inputs_size["x"] == -1
    assert mat1._MatlabDiscipline__outputs_size["y"] == -1
    assert mat1._MatlabDiscipline__is_size_known is False


def test_inputs_and_outputs_size_known_eval():
    """Test that size of input and output variables are known after evaluating matlab
    function."""
    mat2 = MatlabDiscipline(MATLAB_SIMPLE_FUNC_MULTIDIM)
    mat2.execute({"x": array([1, 1]), "y": array([1])})

    assert mat2._MatlabDiscipline__inputs_size["x"] == 2
    assert mat2._MatlabDiscipline__inputs_size["y"] == 1
    assert mat2._MatlabDiscipline__outputs_size["z1"] == 2
    assert mat2._MatlabDiscipline__outputs_size["z2"] == 1
    assert mat2._MatlabDiscipline__is_size_known is True


def test_inputs_and_outputs_size_known_init():
    """Test the size of input and output variables are known when initializing with
    matlab data."""
    mat2 = MatlabDiscipline(
        MATLAB_SIMPLE_FUNC_MULTIDIM, matlab_data_file=FCT_MULTIDIM_DATASET
    )

    assert mat2._MatlabDiscipline__inputs_size["x"] == 2
    assert mat2._MatlabDiscipline__inputs_size["y"] == 1
    assert mat2._MatlabDiscipline__outputs_size["z1"] == 2
    assert mat2._MatlabDiscipline__outputs_size["z2"] == 1
    assert mat2._MatlabDiscipline__is_size_known is True


def test_jac_output_names_error_wrong_name():
    """Test that jacobians output raise an error if names are wrong."""
    with pytest.raises(ValueError) as excp:
        MatlabDiscipline(
            MATLAB_FILES_DIR_PATH / "dummy_test_jac_wrong.m",
            is_jac_returned_by_func=True,
        )
    assert (
        str(excp.value) == "Jacobian terms ['jac_dy_dx'] are not found "
        "in the list of conventional names. "
        "It is reminded that jacobian terms' name "
        "should be such as 'jac_dout_din'"
    )


def test_jac_output_names_error_missing_term():
    """Test that jacobians output raise an error if a term is missing."""
    with pytest.raises(
        ValueError,
        match="The number of jacobian outputs "
        "does not correspond to what it "
        "should be. Make sure that all "
        "outputs have a jacobian matrix "
        "with respect to inputs.",
    ):
        MatlabDiscipline(
            MATLAB_FILES_DIR_PATH / "dummy_test_multidim_jac_wrong.m",
            is_jac_returned_by_func=True,
        )


def test_jac_output_names():
    """Test that jacobians output name are detected when returning by the main
    function."""
    mat = MatlabDiscipline(
        MATLAB_SIMPLE_FUNC_MULTIDIM_JAC, is_jac_returned_by_func=True
    )

    assert mat._MatlabDiscipline__jac_output_names == [
        "jac_dz1_dx",
        "jac_dz1_dy",
        "jac_dz2_dx",
        "jac_dz2_dy",
    ]


def test_jac_output_indices():
    """Test that jacobians output have the right indices."""
    mat = MatlabDiscipline(
        MATLAB_SIMPLE_FUNC_MULTIDIM_JAC, is_jac_returned_by_func=True
    )

    assert mat._MatlabDiscipline__jac_output_indices == [2, 3, 4, 5]


def test_init_default_data():
    """Test that data are correctly initialized."""
    mat = MatlabDiscipline(
        matlab_fct=MATLAB_COMPLEX_FUNC,
        matlab_data_file=MATLAB_FILES_DIR_PATH / "dummy_complex_fct_database.mat",
    )
    assert array(mat.default_inputs["a"]) == pytest.approx(
        array([[[1, 1], [1, 1]], [[1, 1], [1, 1]]])
    )
    assert array(mat.default_inputs["b"]) == pytest.approx(
        array([[[1, 1], [1, 1]], [[1, 1], [1, 1]]])
    )
    assert mat.default_inputs["c"] == pytest.approx(2)
    assert array(mat.default_inputs["d"]) == pytest.approx(array([1, 2]))
    assert mat.default_inputs["e"] == pytest.approx(1)
    assert mat.default_inputs["f"] == pytest.approx(1)


def test_search_file_error_not_found():
    """Test that an error is raised if file is not found."""
    with pytest.raises(
        IOError, match="No file: dummy_test.m, " "found in directory: non_existing."
    ):
        MatlabDiscipline("dummy_test.m", search_file="non_existing")


def test_search_file_error_two_files_found():
    """Test that an error is raised if two files are found."""
    with pytest.raises(IOError) as excp:
        MatlabDiscipline("dummy_test.m", search_file=MATLAB_FILES_DIR_PATH)
    assert (
        str(excp.value) == "At least two files dummy_test.m were "
        "in directory {}\n File one: {};"
        "\n File two: {}.".format(
            MATLAB_FILES_DIR_PATH,
            MATLAB_SIMPLE_FUNC,
            MATLAB_FILES_DIR_PATH / "matlab_files_bis_test" / "dummy_test.m",
        )
    )


def test_search_file():
    """Test that file is found."""
    mat = MatlabDiscipline("dummy_test_multidim.m", search_file=MATLAB_FILES_DIR_PATH)
    assert mat._MatlabDiscipline__inputs == ["x", "y"]


def test_check_existing_function():
    """Test an existing user-made function."""
    mat = MatlabDiscipline(MATLAB_SIMPLE_FUNC)
    assert mat.function_name == "dummy_test"


def test_check_function_builtin():
    """Test a built-in function."""
    mat = MatlabDiscipline("cos", input_names=["x"], output_names=["y"])
    assert mat.function_name == "cos"


def test_run_builtin():
    """Test that built-in matlab function is correctly called and returned right
    values."""
    mat = MatlabDiscipline("cos", input_names=["x"], output_names=["out"])
    mat.execute({"x": array([0])})
    assert mat.local_data["out"] == pytest.approx(1)


def test_run_user():
    """Test that user matlab function is correctly called and returned right values."""
    mat = MatlabDiscipline(MATLAB_SIMPLE_FUNC)
    mat.execute({"x": array([2])})
    assert mat.local_data["y"] == pytest.approx(4)


def test_run_user_new_names():
    """Test that user matlab function is correctly called and returned right values when
    new names are prescribed for inputs and outputs."""
    mat = MatlabDiscipline(
        MATLAB_SIMPLE_FUNC, input_names=["in1"], output_names=["out"]
    )
    mat.execute({"in1": array([3])})
    assert mat.local_data["out"] == pytest.approx(9)


def test_run_user_multidim():
    """Test that user matlab function is correctly called and returned right values."""
    mat = MatlabDiscipline(
        MATLAB_SIMPLE_FUNC_MULTIDIM, matlab_data_file=FCT_MULTIDIM_DATASET
    )
    mat.execute({"x": array([1, 2]), "y": array([3])})
    assert array(mat.local_data["z1"]) == pytest.approx(array([1, 5]))
    assert mat.local_data["z2"] == pytest.approx(11)


def test_run_user_multidim_no_extension_data():
    """Test that an input matlab data file can be prescribed without extension."""
    mat = MatlabDiscipline(
        "dummy_test_multidim.m",
        matlab_data_file="dummy_file_multidim_fct",
        search_file=MATLAB_FILES_DIR_PATH,
    )
    mat.execute({"x": array([1, 2]), "y": array([3])})
    assert array(mat.local_data["z1"]) == pytest.approx(array([1, 5]))
    assert mat.local_data["z2"] == pytest.approx(11)


def test_run_user_multidim_jac():
    """Test that user matlab function is correctly called and returned right values when
    jacobian is defined."""
    mat = MatlabDiscipline(
        MATLAB_SIMPLE_FUNC_MULTIDIM_JAC,
        matlab_data_file=FCT_MULTIDIM_DATASET,
        is_jac_returned_by_func=True,
    )
    mat.execute({"x": array([1, 2]), "y": array([3])})
    assert array(mat.jac["z1"]["x"]) == pytest.approx(array([[2, 3], [0, 4]]))
    assert array(mat.jac["z1"]["y"]) == pytest.approx(array([[0], [54]]))
    assert array(mat.jac["z2"]["x"]) == pytest.approx(array([[4, 0]]))
    assert array(mat.jac["z2"]["y"]) == pytest.approx(array([[6]]))
    assert len(mat.local_data) == 4


def test_run_user_multidim_jac_wrong_size():
    """Test that user matlab function is correctly called and returned right values when
    jacobian is defined."""
    mat = MatlabDiscipline(
        MATLAB_FILES_DIR_PATH / "dummy_test_multidim_jac_wrong_size.m",
        matlab_data_file=FCT_MULTIDIM_DATASET,
        is_jac_returned_by_func=True,
    )
    with pytest.raises(ValueError) as excp:
        mat.execute({"x": array([1, 2]), "y": array([3])})

    assert str(excp.value) == (
        "Jacobian term 'jac_dz1_dx' has the wrong size "
        "(1, 4) whereas it should be (2, 2)."
    )


def test_check_cleaning_interval():
    """Test check that interval cleaning raise an error if not an integer."""
    with pytest.raises(
        ValueError,
        match="The parameter 'cleaning_interval' argument must be an integer.",
    ):
        MatlabDiscipline(MATLAB_SIMPLE_FUNC, clean_cache_each_n=2.3)


def test_save_data():
    """Test that discipline data are correctly exported into a matlab file."""
    mat = MatlabDiscipline(MATLAB_SIMPLE_FUNC)
    mat.execute({"x": array([2])})
    output_file = "output_file.mat"
    mat.save_data_to_matlab(output_file)
    written_data = load_matlab_file(output_file)
    assert array(written_data["x"]) == pytest.approx(2)
    assert array(written_data["y"]) == pytest.approx(4)



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# Copyright (c) 2018 IRT-AESE.
# All rights reserved.
#
# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Arthur Piat
#        :author: François Gallard: initial author of the scilab version
#                                   of MatlabDataProcessorWrapper
#        :author: Nicolas Roussouly: GEMSEO integration
#
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""Definition of Matlab data processor.

Overview
--------

The class and functions in this module enables to
manipulate data from and toward the Matlab workspace.
It also enables to read and write Matlab data file (.mat).
"""
from __future__ import annotations

from copy import copy
from pathlib import Path
from typing import Mapping
from typing import MutableMapping

import matlab
import scipy.io
from gemseo.core.data_processor import DataProcessor
from numpy import array
from numpy import iscomplexobj
from numpy import ndarray


class MatlabDataProcessor(DataProcessor):
    """A Matlab data processor.

    Convert GEMSEO format to Matlab format.

    Examples:
        >>> # Build a new instance
        >>> proc = MatlabDataProcessor()
        >>> # initial python data
        >>> d = {"x": array([2]), "y": array([2j], dtype="complex")}
        >>> # process data to matlab format
        >>> res = proc.pre_process_data(d)
        >>> print(res)
        >>>
        >>> # initial data in matlab format
        >>> d = {"y": double([2, 3]), "x": double([2j], is_complex=True)}
        >>> # process to Python format
        >>> res = proc.post_process_data(d)
        >>> print(res)
    """

    def pre_process_data(
        self, data: Mapping[str, ndarray]
    ) -> Mapping[str, matlab.double]:
        """Transform data from GEMSEO to Matlab.

        The function takes a dict of ndarray and return
        a dict of matlab.double.

        Args:
            data: The input data.

        Returns:
            The data with matlab array types.
        """
        processed_data = {}

        for name, values in data.items():
            if isinstance(values, matlab.double):
                processed_data[name] = values
            else:
                processed_data[name] = array2double(values)

        return processed_data

    def post_process_data(
        self, data: Mapping[str, matlab.double]
    ) -> Mapping[str, ndarray]:
        """Transform the output data from Matlab to GEMSEO.

        Args:
            data: The data with matlab arrays.

        Returns:
            The data with numpy arrays.
        """
        processed_data = {}

        for name, values in data.items():
            if isinstance(values, ndarray):
                processed_data[name] = values.copy()
            else:
                processed_data[name] = double2array(values)

        return processed_data


def load_matlab_file(
    file_path: str | Path,
) -> Mapping[str, matlab.double]:
    """Read .mat file and convert it to usable format for Matlab.

    Args:
        file_path: The path to a .mat file.

    Returns:
        The dict of matlab.double.
    """
    row_data = scipy.io.loadmat(str(file_path))
    clean_data = {}
    for parameter in row_data:
        if parameter not in ["__header__", "__globals__", "__version__"]:
            clean_data[parameter] = array2double(row_data[parameter])

    return clean_data


def save_matlab_file(
    data: MutableMapping[str, ndarray],
    file_path: str | Path = "output_dict",
    **options: bool | str,
) -> None:
    """Save data to a MATLAB-style *.mat* file.

    Args:
        data: The data of the form ``{data_name: data_value}``.
        file_path: The path of the file where to save the data.
        **options: The options of ``scipy.io.savemat``.

    Raises:
        ValueError: If some values in ``data`` are not NumPy arrays.
    """
    data_copy = copy(data)
    for data_name, data_value in data_copy.items():
        if isinstance(data_value, matlab.double):
            data_copy[data_name] = double2array(data_value)
        elif not isinstance(data_value, ndarray):
            raise ValueError("The data must be composed of NumPy arrays only.")
    scipy.io.savemat(str(file_path), data_copy, **options)


def array2double(data_array: ndarray) -> matlab.double:
    """Turn a ndarray into a matlab.double.

    Args:
        data_array: The numpy array to be converted.

    Returns:
        The matlab.double value.
    """
    is_cmplx = iscomplexobj(data_array)
    if len(data_array.shape) == 1:
        return matlab.double(data_array.tolist(), is_complex=is_cmplx)[0]
    else:
        return matlab.double(data_array.tolist(), is_complex=is_cmplx)


def double2array(
    matlab_double: matlab.double,
) -> ndarray:
    """Turn a matlab double into ndarray.

    Args:
        matlab_double: The matlab.double values.

    Returns:
        The array of values.
    """
    if iscomplexobj(matlab_double):
        d_type = "complex"
    else:
        d_type = None

    # note here that we can treat string as well
    # -> we put string into an array as float
    #    (otherwise the array has no shape)
    if isinstance(matlab_double, (float, str, complex)):
        output = array([matlab_double], dtype=d_type)
    else:
        output = array(matlab_double, dtype=d_type)

    if output.shape[0] == 1 and len(output.shape) > 1:
        output = output[0]

    return output


def convert_array_from_matlab(
    data: Mapping[str, matlab.double],
) -> Mapping[str, ndarray]:
    """Convert dict of matlab.output to dict of ndarray.

    Args:
        data: The dict of matlab.double.

    Returns:
        The dict of ndarray.
    """
    output_values = {}
    for matlab_key in data:
        current_value = data[matlab_key]
        output_values[matlab_key] = double2array(current_value)
    return output_values


def convert_array_to_matlab(
    data: Mapping[str, ndarray],
) -> Mapping[str, matlab.double]:
    """Convert gems dict of ndarray to dict of matlab.double.

    Args:
        data: The dict of ndarray.

    Returns:
        The dict of matlab.double.
    """
    output = {}
    for keys in data:
        current_data = data[keys]
        if not (len(current_data) == 1):
            output[keys] = array2double(current_data)
        else:
            if iscomplexobj(current_data):
                output[keys] = complex(current_data)
            else:
                output[keys] = float(current_data)

    return output



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
from __future__ import annotations

import re

import pytest
from gemseo_matlab.matlab_parser import MatlabParser  # noqa: E402

from .matlab_files import MATLAB_FILES_DIR_PATH  # noqa: E402

# TODO: change message because scripts not allowed


@pytest.mark.parametrize(
    "error, match_pattern, path",
    (
        (
            ValueError,
            "The given file {} should either be a matlab function or script.",
            MATLAB_FILES_DIR_PATH,
        ),
        (
            IOError,
            "The function directory for Matlab sources {} does not exists.",
            MATLAB_FILES_DIR_PATH / "not_existing_file.m",
        ),
        (
            ValueError,
            "The given file {} is encrypted and cannot be parsed.",
            MATLAB_FILES_DIR_PATH / "dummy_test.p",
        ),
        (
            ValueError,
            "The given file {} is not a matlab function.",
            MATLAB_FILES_DIR_PATH / "dummy_script.m",
        ),
        (
            ValueError,
            "The given file {} should either be a matlab function or script.",
            MATLAB_FILES_DIR_PATH / "dummy_file.mat",
        ),
        (
            NameError,
            "Function name dummy_test_no_name does not match with file name .",
            MATLAB_FILES_DIR_PATH / "dummy_test_no_name.m",
        ),
        (
            NameError,
            "Function name dummy_test_wrong_name does not match with file name dummy_test.",
            MATLAB_FILES_DIR_PATH / "dummy_test_wrong_name.m",
        ),
        (
            NameError,
            "Matlab function has no name.",
            MATLAB_FILES_DIR_PATH / "dummy_test_no_output.m",
        ),
        (
            NameError,
            "Matlab function has no name.",
            MATLAB_FILES_DIR_PATH / "dummy_test_no_input.m",
        ),
        (
            ValueError,
            "The given file {} is not a matlab function.",
            MATLAB_FILES_DIR_PATH / "not_a_matlab_function.m",
        ),
    ),
)
def test_errors(error, match_pattern, path):
    """Test that exception is raised if file is a directory."""
    # Path on windows that may have regex keywords.
    match = re.escape(match_pattern.format(path))
    with pytest.raises(error, match=match):
        MatlabParser(path)


def test_check_path_good_from_relative_path():
    """Test that file is found when given relative path."""
    parser = MatlabParser(MATLAB_FILES_DIR_PATH / "dummy_test.m")
    assert parser.function_name == "dummy_test"


# TODO: refactor, bad match --> giving a function with no output
#   return the no name exception


# TODO: refactor, bad match --> giving a function with no input
#   return no name exception


def test_scan_onef_complex():
    """Test that a more complex function is correctly parsed."""
    parser = MatlabParser(MATLAB_FILES_DIR_PATH / "dummy_complex_fct.m")
    assert parser.inputs == ["a", "b", "c", "d", "e", "f"]
    assert parser.outputs == ["x", "y", "z"]



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
from __future__ import annotations

import numpy as np
import pytest

# skip if matlab API is not found
pytest.importorskip("matlab")

from gemseo.algos.design_space import DesignSpace  # noqa: E402
from gemseo import create_discipline, create_scenario  # noqa: E402

from .matlab_files import MATLAB_FILES_DIR_PATH  # noqa: E402


def build_matlab_disciplines():
    """Build all matlab discipline for Sellar problem.

    Jacobian matrices are returned by matlab functions.
    """
    matlab_data = MATLAB_FILES_DIR_PATH / "sellar_data.mat"

    sellar1 = create_discipline(
        "MatlabDiscipline",
        matlab_fct="Sellar1.m",
        matlab_data_file=matlab_data,
        name="sellar_1",
        search_file=MATLAB_FILES_DIR_PATH,
        is_jac_returned_by_func=True,
    )

    sellar2 = create_discipline(
        "MatlabDiscipline",
        matlab_fct="Sellar2.m",
        matlab_data_file=matlab_data,
        search_file=MATLAB_FILES_DIR_PATH,
        name="sellar_2",
        is_jac_returned_by_func=True,
    )

    sellar_system = create_discipline(
        "MatlabDiscipline",
        matlab_fct="SellarSystem.m",
        matlab_data_file=matlab_data,
        search_file=MATLAB_FILES_DIR_PATH,
        name="sellar_system",
        is_jac_returned_by_func=True,
    )

    return [sellar1, sellar2, sellar_system]


def build_matlab_scenario():
    """Build the Sellar scenario for matlab tests."""
    design_space = DesignSpace()
    design_space.add_variable("x", l_b=0.0, u_b=10.0, value=np.ones(1))
    design_space.add_variable(
        "z", 2, l_b=(-10, 0.0), u_b=(10.0, 10.0), value=np.array([4.0, 3.0])
    )
    design_space.add_variable("y_1", l_b=-100.0, u_b=100.0, value=np.ones(1))
    design_space.add_variable("y_2", l_b=-100.0, u_b=100.0, value=np.ones(1))

    disciplines = build_matlab_disciplines()
    scenario = create_scenario(
        disciplines, formulation="IDF", objective_name="obj", design_space=design_space
    )

    scenario.add_constraint("c_1", "ineq")
    scenario.add_constraint("c_2", "ineq")

    return scenario


def test_matlab_jacobians_sellar1():
    """Check that jacobian matrices returned by matlab functions are correct with
    respect to finite difference computation for Sellar1."""
    sellar1, sellar2, sellar_system = build_matlab_disciplines()

    threshold = 1e-7
    step = 1e-7

    assert sellar1.check_jacobian(step=step, threshold=threshold)
    assert sellar2.check_jacobian(step=step, threshold=threshold)
    assert sellar_system.check_jacobian(step=step, threshold=threshold)


def test_matlab_optim_results():
    """Test obtained optimal values when solving sellar problem with matlab discipline.

    Jacobians are computed.
    """
    scenario = build_matlab_scenario()
    scenario.execute(input_data={"max_iter": 20, "algo": "SLSQP"})

    # ref values are taken from the doc "Sellar Problem"

    optim_res = scenario.optimization_result
    assert pytest.approx(optim_res.f_opt) == 3.182059

    x_opt = scenario.design_space.get_current_value(as_dict=True)
    assert pytest.approx(x_opt["x"]) == 0.0
    assert pytest.approx(x_opt["z"][0]) == 2.0363869
    assert pytest.approx(x_opt["z"][1]) == 0.0
    assert pytest.approx(x_opt["y_1"]) == 3.16
    assert pytest.approx(x_opt["y_2"]) == 3.814028



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# Copyright (c) 2018 IRT-AESE.
# All rights reserved.
#
# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Arthur Piat
#        :author: Nicolas Roussouly: GEMSEO integration
"""Definition of the Matlab license manager.

Overview
--------

This module contains the :class:`.LicenseManager`
which enables to check the presence of any toolbox licenses
of the Matlab installation.
"""
from __future__ import annotations

import logging
from pathlib import Path
from typing import ClassVar
from typing import Iterable

from gemseo_matlab.engine import MatlabEngine

LOGGER = logging.getLogger(__name__)


class LicenseManager:
    """Manage Matlab License.

    The licenseManager was created to enable de-synchronised
    launch of optimization using matlab_discipline. The goal
    is to wait until all licenses that are needed are available
    in Matlab workspace.
    Parallel computing launch can be used with this tool.

    Examples:
        >>> # Build a new matlab engine
        >>> eng = get_matlab_engine()
        >>> # add a toolbox to the engine
        >>> eng.add_toolbox("signal_toolbox")
        >>> # build a license manager from the previous engine
        >>> lm = LicenseManager(eng)
        >>> # check licenses of the engine until all are available
        >>> lm.check_licenses()
    """

    engine: MatlabEngine
    """The MatlabEngine instance."""

    SIGNAL_TOOL: ClassVar[str] = "signal_toolbox"
    DISTRIB_COMP_TOOL: ClassVar[str] = "distrib_computing_toolbox"
    CURVE_FIT_TOOL: ClassVar[str] = "Curve_Fitting_Toolbox"

    def __init__(self, engine: MatlabEngine) -> None:
        """
        Args:
            engine: The MatlabEngine instance.
        """  # noqa: D205, D212, D415
        self.__engine = engine
        if self.__engine.is_closed:
            # If engine is not (re)started here, add_path
            # will raise an error. A closed engine could happen here
            # since engine is a singleton and could be closed in any
            # other location
            self.__engine.start_engine()

        self.__engine.add_path(str(Path(__file__).parent / "matlab_files"))
        self.licenses = self.__engine.get_toolboxes()

    def check_licenses(
        self,
        licenses: Iterable[str] = (),
        pause_frac: float = 60,
        pause_const: float = 20,
    ) -> None:
        """Check that the Matlab licenses exist.

        The method fetches all the needed licenses thanks to the
        matlab function licenseControl and the class Logger. Note
        that the MATLAB function will be looping until the given
        toolboxes are available.

        Args:
            licenses: The matlab toolboxes.
                If empty, use the already existing engine licenses.
            pause_frac: The time used between each try to get licenses.
            pause_const: The time added in order to estimate the waiting time.
                The waiting time is estimated at each try with the following formula:
                ``Wt = pause_const + random([0,1])*pause_frac``.
        """
        if licenses:
            self.licenses = licenses

        if not self.licenses:
            LOGGER.info("No MATLAB license check will be performed.")
            return

        self.__engine.execute_function(
            "licenseControl", self.licenses, pause_frac, pause_const, nargout=0
        )

    def start_parallel_computing(
        self,
        n_parallel_workers: int = 4,
        cluster_name: MatlabEngine.ParallelType = MatlabEngine.ParallelType.LOCAL,
    ) -> None:
        """Start parallel computing in MatlabEngine.

        Args:
            n_parallel_workers: The number of "workers" to the parallel pool.
                Maximum number allowed is 12
            cluster_name: The matlab parallel pool cluster name.
        """
        self.check_licenses([self.DISTRIB_COMP_TOOL])
        self.__engine.start_parallel_computing(n_parallel_workers, cluster_name)

    def end_parallel_computing(self) -> None:
        """Stop parallel computing in matlab engine."""
        self.__engine.end_parallel_computing()



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# Copyright (c) 2018 IRT-AESE.
# All rights reserved.
#
# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Francois Gallard
#        :author: Arthur Piat
#        :author: Nicolas Roussouly: GEMSEO integration
#
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""Definition of the matlab engine singleton for workspace handling.

Overview
--------

This module contains the :class:`.MatlabEngine` class
which enables to build the Matlab workspace.
The Matlab workspace must be seen as the Matlab "area"
where Matlab functions are executed as well as Matlab variables live.
The engine is basically used when creating a :class:`.MatlabDiscipline` instance
and therefore is not directly handled by the user.
However, a :class:`.MatlabEngine` instance can be used
outside a :class:`.MatlabDiscipline`
in order to directly call Matlab functions
and/or accessing to some variables into the Matlab workspace.

Since :class:`.MatlabEngine` is private, it cannot be used directly from the module.
It is rather used through the function :func:`.get_matlab_engine`
which enables to create only one instance with respect to the ``workspace_name``
(i.e. the instance is unique if the workspace name is the same
when calling several times the function).
Following this, :class:`.MatlabEngine` acts like a singleton.
"""
from __future__ import annotations

import logging
import os
from functools import lru_cache
from pathlib import Path
from typing import Any
from typing import Final

import matlab.engine
from numpy import ndarray  # noqa F401
from strenum import StrEnum

from gemseo_matlab.matlab_data_processor import double2array

LOGGER = logging.getLogger(__name__)


@lru_cache
def get_matlab_engine(
    workspace_name: str = "matlab",
) -> MatlabEngine:
    """Return a new matlab engine.

    LRU cache decorator enables to cache the instance
    if prescribed ``workspace_name`` is the same.
    Therefore, it acts like a singleton.
    This means that calling this function with the same
    ``workspace_name`` returns the same instance.

    Args:
        workspace_name: The name of matlab workspace.

    Returns:
        A Matlab engine instance.

    Examples:
        >>> eng1 = get_matlab_engine()
        >>> eng2 = get_matlab_engine()
        >>> # make sure that engines are the same
        >>> eng1 is eng2
    """
    return MatlabEngine(workspace_name)


class MatlabEngine:
    """Wrapper around the matlab execution engine.

    Since this class is private, an instance should be built through
    :func:`get_matlab_engine` function.
    Using that latter function acts like a singleton,
    i.e. the returned instance is unique if the workspace name is the same.

    Examples:
        >>> eng1 = get_matlab_engine()
        >>> # add new searching directory to workspace with sub-dir
        >>> eng1.add_path("dir_to_files", add_subfolder=True)
        >>>
        >>> # check if a function exists in workspace and returns the type
        >>> exist, type_func = eng1.exist("cos")
        >>> # execute the function
        >>> eng1.execute_function("cos", 0, nargout=1)
        >>>
        >>> # quit engine
        >>> eng1.close_session()
        >>> print(eng1.is_closed)
    """

    class ParallelType(StrEnum):
        """Types of Matlab parallel execution."""

        LOCAL = "local"
        CLOUD = "MATLAB Parallel Cloud"

    __matlab: MatlabEngine
    """The matlab engine."""

    __is_closed: bool
    """Whether the matlab engine session is closed."""

    __paths: list[str]
    """The search paths for the matlab files."""

    __toolboxes: set[str]
    """The names of the matlab toolboxes."""

    __is_parallel: bool
    """Whether the execution is parallel."""

    __EXIST_TO_TYPE_NAME: Final[tuple] = (
        "",
        "variable",
        "file",
        "MEX-file",
        "Simulink-file",
        "MATLAB-built-in-function",
        "P-code-file",
        "folder",
        "class",
    )

    def __init__(self, engine_name: str) -> None:
        """
        Args:
            engine_name: The name identifying the engine.
        """  # noqa: D205, D212, D415
        self.__engine_name = engine_name
        self.__is_closed = True
        self.__paths = []
        self.__toolboxes = set()
        self.__is_parallel = False

        self.start_engine()
        self.add_path(str(Path(__file__).parent))

    @property
    def paths(self) -> list[str]:
        """Return the paths."""
        return self.__paths

    @property
    def is_parallel(self) -> bool:
        """Return True if parallel is active."""
        return self.__is_parallel

    @property
    def is_closed(self) -> bool:
        """Return True if the matlab engine is closed."""
        return self.__is_closed

    def start_engine(self) -> None:
        """Start the matlab engine."""
        LOGGER.info('Starting Matlab engine named "%s".', self.__engine_name)
        self.__matlab = matlab.engine.start_matlab()
        LOGGER.info('Matlab engine named "%s" started', self.__engine_name)

        for path in self.__paths:
            self.__matlab.addpath(path)
        self.__is_closed = False

    def exist(self, name: str | Path) -> tuple[bool, str]:
        """Check if the given matlab file exists.

        Args:
            name: The name to be checked if present in MATLAB path.

        Returns:
            A boolean that tells if ``name`` exist.
            A string that indicates the type of file where
            ``function_name`` is found.
        """
        out = int(self.execute_function("exist", str(name)))
        # 0 — given name does not exist.
        # 1 — name is a variable in the workspace.
        # 2 — name is a file with extension .m, .mlx, or .mlapp,
        #     or name is the name of a file with a non-registered file
        #     extension (.mat, .fig, .txt).
        # 3 — name is a MEX-file on your MATLAB search path.
        # 4 — name is a loaded Simulink® model or a Simulink model or
        #     library file on your MATLAB search path.
        # 5 — name is a built-in MATLAB function. This does not include classes.
        # 6 — name is a P-code file on your MATLAB search path.
        # 7 — name is a folder.
        # 8 — name is a class. (exist returns 0 for Java classes if
        #     you start MATLAB with the -nojvm option.)
        return out != 0, self.__EXIST_TO_TYPE_NAME[out]

    def close_session(self) -> None:
        """Close the matlab session."""
        self.__matlab.quit()
        self.__is_closed = True

    def add_toolbox(self, toolbox_name: str) -> None:
        """Add a toolbox to the engine.

        The toolbox added would be needed for the functions
        used in the current session. It should be checked that
        the license is compatible.
        The name given here can be found using "license('inuse')" in MATLAB.

        Args:
            toolbox_name: The name of the toolbox to be checked.
        """
        self.__toolboxes.add(toolbox_name)

    def remove_toolbox(self, toolbox_name: str) -> None:
        """Remove a toolbox from the engine.

        Args:
            toolbox_name: The name of the toolbox to be checked.
        """
        self.__toolboxes.remove(toolbox_name)

    def add_path(
        self,
        path: str | Path,
        add_subfolder: bool = False,
    ) -> None:
        """Add a path to the matlab engine search directories.

        Args:
            path: The path to the directory or file to be added to path.
            add_subfolder: Whether to add the sub-folders recursively.

        Raises:
            ValueError: If the given path cannot be added to Matlab.
        """
        path = str(path)
        if path not in self.__paths:
            if not self.exist(path):
                raise ValueError("The given path cannot be added to matlab.")
            self.__paths.append(path)
            self.__matlab.addpath(path)

        if add_subfolder:
            str_paths = self.execute_function("genpath", path)
            # first path in list is the path already in list
            for sub_path in str_paths.split(os.pathsep)[1:]:
                self.add_path(sub_path)

    def execute_function(
        self,
        func_name: str,
        *args: float | str,
        **kwargs: float,
    ) -> Any:
        """Executes a Matlab function called "func_name".

        Args:
            func_name: The function name to call.
            *args: Any arguments that must be passed to the function.
            **kwargs: Any arguments that must be passed to the function.

        Raises:
            matlab.engine.MatlabExecutionError: If the matlab function execution fails.
        """
        if self.__is_closed:
            self.start_engine()

        method = getattr(self.__matlab, func_name)

        try:
            return method(*args, **kwargs)

        except matlab.engine.MatlabExecutionError:
            LOGGER.error(
                "Failed to execute Matlab function %s with arguments %s and %s",
                func_name,
                str(args),
                str(kwargs),
            )
            raise

    def start_parallel_computing(
        self,
        n_parallel_workers: int,
        parallel_type: ParallelType = ParallelType.LOCAL,
    ) -> None:
        """Start the parallel pool of matlab for parallel computing.

        This feature only works if parallel toolbox is available.

        Args:
            n_parallel_workers: The number of "workers" to the parallel pool.
            parallel_type: The type of parallel execution.
        """
        exist, _ = self.exist("parpool")

        if exist is False:
            LOGGER.warning("Parallel computing not available in MATLAB")
            self.__is_parallel = False
            return

        LOGGER.info(
            "Starting parallel computing with %s workers on '%s' cluster.",
            n_parallel_workers,
            parallel_type,
        )

        try:
            self.execute_function(
                "parpool", parallel_type, float(n_parallel_workers), nargout=0
            )
        except matlab.engine.MatlabExecutionError:
            LOGGER.warning(
                "Parallel computing could not be started, proceeding without."
            )
            self.__is_parallel = False
        else:
            self.__is_parallel = True

            LOGGER.info("MATLAB parallel computing successfully started.")
            LOGGER.info(
                "Only MATLAB function and classes using some 'parfor' will be "
                "computed in parallel."
            )

    def end_parallel_computing(self) -> None:
        """End the parallel computing.

        Raises:
            matlab.engine.MatlabExecutionError: If the parallel option is not
                correctly deactivated.
        """
        if not self.__is_parallel:
            LOGGER.info("Try to end parallel computing whereas it is not activated.")
            return

        LOGGER.info("Closing MATLAB parallel pools.")

        try:
            self.execute_function("eval", "delete(gcp('nocreate'))", nargout=0)
        except matlab.engine.MatlabExecutionError:
            LOGGER.warning("Parallel computing could not be closed.")
            raise
        else:
            self.__is_parallel = False

    def get_toolboxes(self) -> set[str]:
        """Return all toolboxes to be checked before launching this engine.

        Returns:
            All toolboxes.
        """
        return self.__toolboxes

    def execute_script(self, script_name: str) -> None:
        """Execute a script in the current workspace.

        After executing the script, the workspace point to the path
        where the script is located.

        Args:
            script_name: The script name.
        """
        # get the script absolute path
        entire_path = self.execute_function("which", script_name, nargout=1)
        abspath, _, _ = self.execute_function("fileparts", entire_path, nargout=3)

        # move to the script path
        self.execute_function("cd", abspath, nargout=0)

        # execute the script
        self.execute_function("run", script_name, nargout=0)

    def get_variable(self, item: str) -> ndarray:
        """Get any variable in the workspace.

        Args:
            item: The variable name.

        Returns:
            The value of the variable.

        Raises:
            ValueError: If the item is unknown inside the workspace.
        """
        try:
            return double2array(self.__matlab.workspace[item])
        except matlab.engine.MatlabExecutionError:
            raise ValueError(
                f"The variable {item} does not exist in the "
                f"current {self.__engine_name} workspace."
            )

    def __del__(self) -> None:
        self.close_session()



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
from __future__ import annotations

import os

import matlab
import pytest
from gemseo_matlab.matlab_data_processor import array2double
from gemseo_matlab.matlab_data_processor import convert_array_from_matlab
from gemseo_matlab.matlab_data_processor import convert_array_to_matlab
from gemseo_matlab.matlab_data_processor import double2array
from gemseo_matlab.matlab_data_processor import load_matlab_file
from gemseo_matlab.matlab_data_processor import MatlabDataProcessor
from gemseo_matlab.matlab_data_processor import save_matlab_file
from numpy import array
from numpy import ndarray

from .matlab_files import MATLAB_FILES_DIR_PATH  # noqa: E402


def test_pre_process_data():
    """Test that data are correctly pre-processed from dict of array to dict of
    matlab.double."""
    data_proc = MatlabDataProcessor()
    d = {"x": array([2]), "y": array([2j], dtype="complex")}
    res = data_proc.pre_process_data(d)
    assert isinstance(res["x"], matlab.double)
    assert res["x"][0] == pytest.approx(2)
    assert isinstance(res["y"], matlab.double)
    assert res["y"][0] == pytest.approx(2j)


def test_post_process_data():
    """Test that data are correctly post-processed from dict of matlab.double to dict of
    array."""
    data_proc = MatlabDataProcessor()
    d = {"y": matlab.double([2, 3]), "x": array([2j], dtype="complex")}
    res = data_proc.post_process_data(d)

    assert isinstance(res["x"], ndarray)
    assert res["x"][0] == pytest.approx(2j)
    assert isinstance(res["y"], ndarray)
    assert res["y"][0] == pytest.approx(2)
    assert res["y"][1] == pytest.approx(3)


def test_load_matlab_file():
    """Test matlab file loading."""
    res = load_matlab_file(MATLAB_FILES_DIR_PATH / "dummy_file.mat")
    assert array(res["a"]) == pytest.approx(2.0)
    assert array(res["b"]) == pytest.approx(3.0)
    assert array(res["c"]) == pytest.approx(4.0)
    assert array(res["d"]) == pytest.approx(array([[4.0, 5.0, 6.0]]))
    assert array(res["e"]) == pytest.approx(array([[10], [20], [30]]))
    assert array(res["f"]) == pytest.approx(array([[1, 2, 3], [4, 5, 6], [7, 8, 99]]))


def test_save_matlab_file():
    """Test matlab file saving."""
    name_file = "test_file.mat"
    test_dict = {
        "test1": array([3.0 + 1j, 2.0], dtype="complex"),
        "test2": array([3.0]),
        "test3": array([[3.0, 2.0], [10.0, 22.0]]),
    }
    save_matlab_file(test_dict, name_file)
    load_dat = load_matlab_file(name_file)

    assert array(load_dat["test1"]).flatten() == pytest.approx(test_dict["test1"])
    assert array(load_dat["test2"]).flatten() == pytest.approx(test_dict["test2"])
    assert array(load_dat["test3"]) == pytest.approx(test_dict["test3"])
    os.remove(name_file)


def test_array2double_1d():
    """Test that a 1d array is correctly converted to double."""
    a = array([3.0, 2.0])
    r = array2double(a)
    assert a == pytest.approx(r)


def test_array2double_1d_complex():
    """Test that a 1d complex array is correctly converted to double."""
    a = array([3.0, 2.0 + 5.2j])
    r = array2double(a)
    assert a == pytest.approx(r)


def test_array2double_2d():
    """Test that a 2d array is correctly converted to double."""
    a = array([[3.0, 2.0], [11, 22.0]])
    r = array2double(a)
    assert a - r == pytest.approx(0)


def test_double2array_1d():
    """Test that matlab.double is correctly converted to ndarray."""
    a = matlab.double([2.0, 3.0])
    r = double2array(a)
    assert a - r == pytest.approx(0)


def test_double2array_1d_complex():
    """Test that 1d complex matlab.double is correctly converted to ndarray."""
    a = matlab.double([2.0, 3.0 + 6.3j], is_complex=True)
    r = double2array(a)
    assert a - r == pytest.approx(0)


def test_double2array_2d():
    """Test that 2d matlab.double is correctly converted to ndarray."""
    a = matlab.double([[3.0, 2.0], [11, 22.0]])
    r = double2array(a)
    assert a - r == pytest.approx(0)


def test_matlab2gems():
    """Test that matlab dict is correctly converted to gems-like dict."""
    d = {
        "test1": matlab.double([3.0 + 1j, 2.0], is_complex=True),
        "test2": matlab.double([3.0]),
        "test3": matlab.double([[3.0, 2.0], [10.0, 22.0]]),
    }
    r = convert_array_from_matlab(d)
    assert r["test1"] - d["test1"] == pytest.approx(0)
    assert r["test2"] - d["test2"] == pytest.approx(0)
    assert sum(r["test3"] - d["test3"]) == pytest.approx(0)


def test_gems2matlab():
    """Test that gems-like dict is correctly converted to matlab dict."""
    d = {
        "test1": array([3.0 + 1j, 2.0]),
        "test2": array([3.0]),
        "test3": array([[3.0, 2.0], [10.0, 22.0]]),
    }
    r = convert_array_to_matlab(d)
    assert sum(r["test1"] - d["test1"]) == pytest.approx(0)
    assert r["test2"] - d["test2"] == pytest.approx(0)
    assert sum(r["test3"] - d["test3"]) == pytest.approx(0)



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
from __future__ import annotations

from pathlib import Path

import matlab
import numpy as np
import pytest
from gemseo_matlab.engine import get_matlab_engine  # noqa: E402

from .matlab_files import MATLAB_FILES_DIR_PATH  # noqa: E402


@pytest.fixture(scope="module")
def matlab_engine():
    """Return a brand new matlab engine with clean cache."""
    get_matlab_engine.cache_clear()
    matlab = get_matlab_engine()
    matlab.add_path(MATLAB_FILES_DIR_PATH, True)
    yield matlab
    get_matlab_engine.cache_clear()


def test_singleton():
    """Test that the engine is effectively a singleton."""
    assert get_matlab_engine() is get_matlab_engine()


def test_start_engine(matlab_engine):
    """Test that matlab engine starts correctly."""
    assert matlab_engine.is_closed is False


def execute_function(matlab_engine):
    matlab_engine.execute_function("assignin", "base", "dummy_var", 2, nargout=0)


@pytest.mark.parametrize(
    "setup, name, is_found, type_",
    (
        (None, "sin", True, "MATLAB-built-in-function"),
        (None, MATLAB_FILES_DIR_PATH, True, "folder"),
        (None, "dummy_test.m", True, "file"),
        (None, "dummy_test.p", True, "P-code-file"),
        (None, "non_existing_file", False, ""),
        (execute_function, "dummy_var", True, "variable"),
    ),
)
def test_exist(setup, name, is_found, type_):
    """Test the existence of a built-in functions."""
    matlab_engine = get_matlab_engine()
    matlab_engine.add_path(MATLAB_FILES_DIR_PATH, True)
    if setup is not None:
        setup(matlab_engine)
    assert matlab_engine.exist(name) == (is_found, type_)


# TODO: add test on MEX-file existence
# TODO: add test on Simulink-file existence
# TODO: add test on class existence


def test_add_toolbox(matlab_engine):
    """Test an add of toolbox."""
    matlab_engine.add_toolbox("signal_toolbox")
    assert "signal_toolbox" in matlab_engine.get_toolboxes()


def test_remove_toolbox(matlab_engine):
    """Test a remove of toolbox."""
    matlab_engine.add_toolbox("signal_toolbox")
    matlab_engine.remove_toolbox("signal_toolbox")
    assert "signal_toolbox" not in matlab_engine.get_toolboxes()


def test_add_path_sub_folder(matlab_engine):
    """Test that sub-folder is correctly added in path."""
    assert str(MATLAB_FILES_DIR_PATH) in matlab_engine.paths
    assert str(MATLAB_FILES_DIR_PATH / "matlab_files_bis_test") in matlab_engine.paths


def test_execute_error(matlab_engine):
    """Test that execution raise an error if something wrong."""
    with pytest.raises(matlab.engine.MatlabExecutionError):
        matlab_engine.execute_function("dummy_test")


@pytest.mark.parametrize("x, res", [(2, 4), (3, 9), (4, 16)])
def test_execute(x, res, matlab_engine):
    """Test the execution of a matlab function."""
    assert res == matlab_engine.execute_function("dummy_test", x, nargout=1)


def test_parallel(matlab_engine):
    """If parallel exist, test that it starts and stop correctly."""
    matlab_engine.start_parallel_computing(4)
    assert matlab_engine.is_parallel is matlab_engine.exist("parpool")[0]

    matlab_engine.end_parallel_computing()
    assert matlab_engine.is_parallel is False


def test_launch_script_dir(matlab_engine):
    """Test script launching and make sure that the path of matlab workspace is the
    directory that contains the script."""
    matlab_engine.execute_script("dummy_script_2")
    path = matlab_engine.execute_function("pwd", nargout=1)
    assert (
        Path(path).absolute()
        == (MATLAB_FILES_DIR_PATH / "matlab_files_bis_test").absolute()
    )


def test_get_variable(matlab_engine):
    """Test getting variable from the matlab workspace to Python."""
    matlab_engine.execute_script("dummy_script_2")
    assert matlab_engine.get_variable("a") == pytest.approx(2)
    assert matlab_engine.get_variable("b") == pytest.approx(3)
    assert matlab_engine.get_variable("c") == pytest.approx(np.array([10, 20]))
    assert matlab_engine.get_variable("d") == np.array(["test_string"])


def test_get_variable_error(matlab_engine):
    """Test that an error is raised if we try to get a variable that does not exist in
    matlab workspace."""
    with pytest.raises(
        ValueError,
        match="The variable toto does not exist in the current matlab workspace.",
    ):
        matlab_engine.get_variable("toto")


def test_quit(matlab_engine):
    """Test that engine is correctly quit."""
    matlab_engine.close_session()
    assert matlab_engine.is_closed is True



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
from __future__ import annotations

import pytest
from gemseo_matlab.engine import get_matlab_engine  # noqa: E402
from gemseo_matlab.license_manager import LicenseManager  # noqa: E402


@pytest.fixture(scope="module")
def matlab_engine():
    """Return a brand new matlab engine with clean cache."""
    get_matlab_engine.cache_clear()
    yield get_matlab_engine()
    get_matlab_engine.cache_clear()


@pytest.mark.parametrize(
    "toolbox", (LicenseManager.CURVE_FIT_TOOL, LicenseManager.DISTRIB_COMP_TOOL)
)
def test_check_when_adding_toolbox_curve_fit(matlab_engine, toolbox):
    """Check that the license is checked when curve fit toolbox is added in engine."""
    matlab_engine.add_toolbox(toolbox)
    manager = LicenseManager(matlab_engine)
    manager.check_licenses()
    assert manager.licenses == matlab_engine.get_toolboxes()
